/* YAWN, Yet Another Web Navigator */
/* Copyright (C) 1998, James Troup <jjtroup@comp.brad.ac.uk>
 *                     Lee Willis <lwillis@comp.brad.ac.uk> */

/* This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include "attrib_utils.h"
#include "parse_attribs.h"
#include "utils.h"
#include "tagdefs.h"
#include "neve.h"
#include "staxx.h"
#include "smg.h"
#include "url.h"

/* parse_attribs.c deals with the parsing of the attributes within HTML tags
 * it also decides whether the information provided makes sense within the
 * current tag context and if so where it should put the resulting information
 * 
 * Each function documented seperately
 */

void handle_src (const char *document, long *index, neve_in_chains *node, const int tag)
{
  /* Parses src="foo bar" and src="foobar" constructs and decides what
   * to do with them. ie src makes no sense inside an <BASEFONT> tag so we
   * ignore it. Only makes sense inside an anchor tag.
   */
  int quoted = 0;
  int src_counter = 0;
  
  debug (debug_parse_attribs, "handle_src: entering pointing at %ld:%s...", 
	 *index, get_bit (document, index) ? foo : "NULL");
  
  (*index)+=3; /* Skip SRC */
  
  if (skip_equals(document, index) == -1)
    return;

  if (document[*index]=='"') 
    { /* Recognize if we're href="foo vr" type and skip
       * opening " */
      debug (debug_parse_attribs, "handle_src: found \".");
      quoted = 1;
      (*index)++;
    }
  
  /* Count how long the src information is. Terminated by a " if we're
   * parsing a quoted form or a ' ' or '>' if we're not
   */
  while (((document[*index] != ' ' && document[*index] != '>') && !quoted) || 
	 (document[*index] != '"' && quoted))
    {
      src_counter++;
      (*index)++;
    }

  debug (debug_parse_attribs, "handle_src: src information is %d characters.", src_counter);
  
  /* Create space for the src information and copy it into the current node */
  if (tag == ANCHOR || tag == IMAGE)
    {
      debug (debug_parse_attribs, "handle_src: Copying into node.");
      node->src = xmalloc (src_counter+1);
      strncpy(node->src, document+*index-src_counter, src_counter);
      node->src[src_counter] = '\0';
    }
  
  /* Skip the close quote if we're present.
   * NB this doesn't check the quoted flag so that we can successfully 
   * deal with <IMG src="foobar.gif> without falling over.
   */
  if (document[*index]=='"') 
    (*index)++;

  /* Skip whitespace until end of tag */
  while ((is_whitespace(document[*index])) && document[*index] != '>')
    (*index)++;

  debug (debug_parse_attribs, "handle_src: Leaving pointing at %ld:%s...", 
	 *index, get_bit (document, index) ? foo : "NULL");
}



void handle_href (const char *document, long *index, int tag, char *current_href, BrowserWindow *window)
{
  /* Parses href="foo bar" and href="foobar" constructs and decides what
   * to do with them.  Essentially the same as handle_src (Above.)
   */
  int quoted = 0;
  int href_counter = 0;
  
  debug (debug_parse_attribs, "handle_href: Entering pointing at %ld:%s...",
	 *index, get_bit (document, index) ? foo : "NULL");
  (*index)+=4; /* Skip HREF */
  debug (debug_parse_attribs, "handle_href: Skipped HREF, now pointing at %ld:%s...",
	 *index, get_bit (document, index) ? foo : "NULL");
  
  if (skip_equals(document, index) == -1)
    return;
  
  if (document[*index]=='"') 
    {
      debug (debug_parse_attribs, "handle_href: Value is quoted.");
      quoted = 1;
      (*index)++;
    }
  
  /* Skip until we find a space or end of tag > if we're not quoted or until we
   * find a " if we are quoted */
  while (((document[*index] != ' ' && document[*index] != '>') && !quoted) || 
	 (document[*index] != '"' && quoted))
    {
      href_counter++;
      (*index)++;
    }
  
  debug (debug_parse_attribs, "handle_href: HREF info is %d characters.", 
	 href_counter);

  debug (debug_parse_attribs, "handle_href: tag is %d", tag);
  if (tag == ANCHOR || tag == IMAGE) 
    {
      strncpy(current_href, document+*index-href_counter, href_counter);
      current_href[href_counter] = '\0';
      absolutize_url (current_href, window);
      debug (debug_parse_attribs, "handle_href: HREF info is %s", 
	     current_href ? current_href : "NULL");
    }
  
  if (document[*index]=='"')
    (*index)++;

  /* Skip whitespace until end of tag */
  while ((is_whitespace(document[*index])) && document[*index] != '>')
    (*index)++;
  
  debug (debug_parse_attribs, "handle_href: Leaving pointing at %ld:%s...", 
	 *index, get_bit(document, index) ? foo : "NULL");

}




void handle_align (const char *document, long *index, neve_in_chains *node, int tag,
		   physical_attribute_types *pattribs)
{
  /* Parse align=center/left/right/top/middle/bottom for headers, paragraph,
   * horizontal rule and image tags.
   * NB. Some of these don't make sense according to the standard so we have
   * to be careful about if we copy the alignment over
   */
  debug (debug_parse_attribs, "handle_align: Entering pointing at %ld:%s...", 
	 *index, get_bit (document, index) ? foo : "NULL");
  (*index)+=5; /* Skip ALIGN */
  debug (debug_parse_attribs, "handle_align: Skipped ALIGN, now pointing at %ld:%s...", 
	 *index, get_bit (document, index) ? foo : "NULL");
  if (skip_equals(document, index) == -1)
    return;
  /* Yark should we check for quotes? I think so, it's not in the standard but
   * hey people consistently write daft HTML so we'll check it just in case
   */
  if (document[*index] == '"')
    {
      debug (debug_parse_attribs, "handle_align: Found open quote, skipping.");
      (*index)++;
    }
  if ((tag >= H1 && tag <=H6) || tag==PARAGRAPH || tag==DIV || 
      tag==HRULE || tag == IMAGE)
    {
      if (strncasecmp(&(document[*index]), "left", 4) == 0)
	{
	  /* Allowable for all the above cases */
	  debug (debug_parse_attribs, "handle_align: left found.");
	  if (tag == IMAGE || tag == HRULE)
	    node->physical_attributes.alignment = left;
	  else
	    pattribs->alignment = left;
	  (*index) += 4;
	}
      else if (strncasecmp(&(document[*index]), "right", 5) == 0)
	{
	  /* Allowable for all the above cases */
	  debug (debug_parse_attribs, "handle_align: right found.");
	  if (tag == IMAGE || tag == HRULE)
	    node->physical_attributes.alignment = right;
	  else
	    pattribs->alignment = right;
	  (*index) += 5;
	}
      else if (strncasecmp(&(document[*index]), "center", 6) == 0)
	{
	  debug (debug_parse_attribs, "handle_align: center found.");
	  if (tag != IMAGE) /* IMG isn't supposed to have an align=center */
	    if (tag == HRULE)
	      node->physical_attributes.alignment = centre;
	    else
	      pattribs->alignment = centre;
	  else
	    debug ( debug_parse_attribs, "Ignored.");
	  (*index) += 6;
	}
      else if (strncasecmp(&(document[*index]), "bottom", 6) == 0)
	{
	  debug (debug_parse_attribs, "handle_align: bottom found.");
	  if (tag == IMAGE) /* bottom, middle and top are only applicable to
			   * IMG tags. */
	    node->physical_attributes.alignment = bottom;
	  else
	    debug ( debug_parse_attribs, "Ignored.");
	  (*index) += 6;
	}
      else if (strncasecmp(&(document[*index]), "middle", 6) == 0)
	{
	  debug (debug_parse_attribs, "handle_align: middle found.");
	  if (tag == IMAGE)
	    node->physical_attributes.alignment = middle;
	  else
	    debug (debug_parse_attribs, "Ignored.");
	  (*index) += 6;
	}
      else if (strncasecmp(&(document[*index]), "top", 3) == 0)
	{
	  debug (debug_parse_attribs, "handle_align: top found.");
	  if (tag == IMAGE)
	    node->physical_attributes.alignment = top;
	  else
	    debug (debug_parse_attribs, "Ignored.");
	  (*index) += 3;
	}
    }
  if (document[*index] == '"')
    (*index)++;
  /* Skip whitespace until end of tag */
  while ((is_whitespace(document[*index])) && document[*index] != '>')
    (*index)++;
  debug (debug_parse_attribs, "handle_align: Leaving pointing at %ld:%s...", 
	 *index, get_bit (document, index) ? foo : "NULL");
}




void handle_unrecognized (const char *document, long *index)
{
  /* Skip an attribute if we don't recognize it. We'll assume 
   * that it goes name[ ]{=[ ]value} */
  
  while ((!(is_whitespace(document[*index])) && 
	  (document[*index] != '=') &&
	  (document[*index] != '>')))
    (*index)++;

  while (is_whitespace(document[*index]) ) /* Skip whitespace */
    (*index)++;

  if ( document[*index]=='=' ) 
    { /* If it has an assignment part then skip that ...*/
      (*index)++; /* Skip the = */
      while (is_whitespace(document[*index])) 
	/* Skip optional whitespace after the = but before the value */
	(*index)++;
      while (!(is_whitespace(document[*index])) && document[*index] != '>')
	/* Skip the value, assume no spaces */
	(*index)++;
    }

}



void handle_size (const char *document, long *index,neve_in_chains *node, const int tag, 
		  int *basefont, boolean *font_change_implemented, 
		  physical_attribute_types *current_pattributes)
{     
  /* Handle the size=dd, size="dd", size=+1, size ="+1" cases
   * This is not nice to handle, but I'm pretty sure this works now
   */

  int size = *basefont;  /* Default the size to basefont so when we get
			  * a relative size it adds on correctly */
  (*index)+=4; /* Skip SIZE */
  if (skip_equals(document, index) == -1)
    return;
  debug (debug_parse_attribs,"handle_size: Checking for quoted values.");
  if (document[*index] == '"')
    {
      (*index)++;
      debug (debug_parse_attribs, "handle_size: \" found.");
    }
  if ((document[*index] == '+' || document[*index] == '-'))
    {
      debug ( debug_parse_attribs,"Found a relative value %c", document[*index]);
      if (tag == FONT || tag == HRULE)
	if (document[*index] == '-')
	  {
	    (*index)++; /* Skip - */
	    size = *basefont - handle_value (document, index);
	    debug (debug_parse_attribs, "handle_size: size now equals %d.", size);
	  }
	else 
	  {
	    (*index)++; /* Skip + */
	    size = *basefont + handle_value (document, index);
	    debug (debug_parse_attribs, "handle_size: size now equals %d.", size);
	  }
      else
	{
	  handle_value (document, index); /* Get the value but ignore it */
	  debug (debug_parse_attribs, 
		 "handle_size: ignoring value, relative value makes no sense in tag type %d",
		 tag);
	}
    }
  else /* Surely we don't need this -> if (document[*index] != '+' && document[*index] != '-') */
    {
      size = handle_value (document, index);
      debug (debug_parse_attribs, "handle_size: size now equals %d.", size);
    }
  if (tag == FONT || tag == BASEFONT) 
    {
      if (size  > 7)  /* Make sure the size is a valid range for the FONT */
	size = 7;     /* and BASEFONT tags. */
      else if (size < 1)
	size = 1;
      if (tag == FONT)
	{
	  if (!(*font_change_implemented))
	    {
	      debug(debug_parse_attribs, "handle_size: creating new font stack node.");
	      push_font (current_pattributes->draw_colour, current_pattributes->font_level);
	      *font_change_implemented = TRUE;
	    }
	  current_pattributes->font_level = size;
	}
      if (tag == BASEFONT)
	*basefont = size;
    }
  else if (tag == HRULE) 
    node->physical_attributes.height = size;  /* Place the correct size into our node
						 * NB, don't bounds check or handle % values
						 * correctly, well at all to be perfectly
						 * honest */
  if (document[*index] == '"')
    (*index)++;
}



void handle_colour(const char *document, long *index, const int tag, 
		   boolean *font_change_implemented, 
		   physical_attribute_types *current_pattributes)
{
  /* parse the color="#xxxxxx" structure, should also parse the common
   * color names, but erm doesn't yet.  
   */
  colour_type value = {-1,-1,-1};
  
  (*index)+=5; /* Skip COLOR */
  if (skip_equals (document, index) == -1)
    return;
  if (document[*index] == '"')
    (*index)++;
  if (document[*index] == '#') 
    {
      (*index)++; /* Skip the # */
      if ((value.red = convert_hex (document, index)) == -1)
	return;
      debug ( debug_parse_attribs, "handle_colour: Red value = %d", value.red);  
      if ((value.green = convert_hex (document, index)) == -1)
	return;
      debug ( debug_parse_attribs, "handle_colour: Green value = %d", value.green);
      if ((value.blue = convert_hex (document, index)) == -1)
	return;
      debug ( debug_parse_attribs, "handle_colour: Blue value = %d", value.blue);
      if (document[*index] =='"')
	(*index)++;
      if (tag == FONT) /* Doesn't make any sense for any other tag */
	{
	  if (!(*font_change_implemented))
	    {
	      debug ( debug_parse_attribs, 
		      "handle_colour: Creating new font stack node.");
	      push_font (current_pattributes->draw_colour, current_pattributes->font_level); /* Create a new font stack node */
	    }
	  current_pattributes->draw_colour = value;
	}
    }
  else
    {
      ; /* Should check the named colors here */
    }
}



void handle_alt (const char *document, long *index, neve_in_chains *node, int tag)
{
  /* handles the ALT=alttext and ALT="alt text" attributes for the IMG tag,
   * NB Makes no sense to anything else.
   */
  
  int quoted = 0;
  long buf_counter = 0;
  
  debug (debug_parse_attribs, "handle_alt: Entering pointing at %ld:%s...",
	 *index, get_bit (document, index)  ? foo : "NULL" );
  (*index) += 3; /* Skip ALT */
  debug (debug_parse_attribs, "handle_alt: Skipped ALT, now pointing at %ld:%s",
	 *index, get_bit (document, index)  ? foo : "NULL" );
  if (skip_equals (document, index) == -1)
    return;
  if (document[*index]=='"')
    {
      quoted = 1;
      (*index)++;
    }
  /* Skip until we hit whitespace or end of tag if we're unquoted or
   * until we find a close quote if we're not quoted.
   */
  while (((document[*index] != ' ' && document[*index] != '>') && !quoted) || 
	 ((document[*index] != '"') && quoted))
    {
      buf_counter++;
      (*index)++;
    }
  debug (debug_parse_attribs, "handle_alt: ALT info is %ld characters long.",
	 buf_counter);
  if (tag == IMAGE) /* Makes no sense for anything else */
    {
      node->buf = xmalloc(buf_counter+1);
      strncpy(node->buf, document+*index-buf_counter, buf_counter);
      node->buf[buf_counter] ='\0';
      debug (debug_parse_attribs, "handle_alt: Info is %s", node->buf);
    }
  if (document[*index]=='"')
    (*index)++;
  debug (debug_parse_attribs, "handle_alt: Leaving pointing at %ld:%s...",
	 *index, get_bit (document, index)  ? foo : "NULL" );
}


void handle_type (const char *document, long *index, logical_attribute_types *lattribs, int tag)
{
  /* Handle the type = disc/square/circle/1/a/A/i/I for determining the
   * list bullet types or numbering type
   */

  debug ( debug_parse_attribs, "handle_type: Entering pointing at %ld:%s...",
	  *index, get_bit(document, index) ? foo : "NULL");
  (*index) += 4;
  debug ( debug_parse_attribs, "handle_type: Skipped TYPE, now pointing at %ld:%s...",
	  *index, get_bit(document, index) ? foo : "NULL");
  if (skip_equals (document, index) == -1 )
    return;
  if (tag == OLIST) /* The following make no sense in any other tag */
    {
      if (document[*index] == '1')
	{
	  debug (debug_parse_attribs, "handle_type: Set type to 1,2,3 etc.");
	  lattribs->ol_number_type = arabic;
	  (*index)++;
	}
      else if (document[*index] == 'a')
	{
	  debug (debug_parse_attribs, "handle_type: Set type to a,b,c etc.");
	  lattribs->ol_number_type = lalpha;
	  (*index)++;
	}
      else if (document[*index] == 'A')
	{
	  debug (debug_parse_attribs, "handle_type: Set type to A,B,C etc.");
	  lattribs->ol_number_type = ualpha;
	  (*index)++;
	}
      else if (document[*index] == 'i')
	{
	  debug (debug_parse_attribs, "handle_type: Set type to i,ii,iii etc.");
	  lattribs->ol_number_type = lroman;
	  (*index)++;
	}
      else if (document[*index] == 'I')
	{
	  debug (debug_parse_attribs, "handle_type: Set type to I, II, III etc.");
	  lattribs->ol_number_type = uroman;
	  (*index)++;
	}
      else 
	{
	  debug (debug_parse_attribs, "handle_type: defaulting to 1,2,3 etc.");
	  lattribs->ol_number_type = 1; /* Default to 1,2,3 */
	  while (document[*index] && (!(is_whitespace(document[*index]))) && document[*index] != '>')
	    (*index)++;
	}
    }
  else if (tag == ULIST) /* The following make no sense in any other tag */
    {
      if (strncasecmp(document+*index, "disc", 4) == 0)
	{
	  debug (debug_parse_attribs, "handle_type: bullet type is now disc.");
	  lattribs->ul_bullet_type = disc;
	  (*index)+=4;
	}
      else if (strncasecmp(document+*index, "square", 6) ==0)
	{
	  debug (debug_parse_attribs, "handle_type: bullet type is now square.");
	  lattribs->ul_bullet_type = square;
	  (*index) += 6;
	}
      else if (strncasecmp(document+*index, "circle", 6) ==0)
	{
	  debug (debug_parse_attribs, "handle_type: bullet type is now circle.");
	  lattribs->ul_bullet_type = circle;
	  (*index) += 6;
	}
      else
	{
	  debug (debug_parse_attribs, "handle_type: defaulting to disc.");
	  lattribs->ul_bullet_type = 1; /* Default to disc */
	  while (document[*index] && (!(is_whitespace(document[*index]))) && document[*index] != '>')
	    (*index)++;
	} 
    }
  while (document[*index] && is_whitespace(document[*index]))
    (*index)++;
}


void handle_width (const char *document, long *index, neve_in_chains *node, int tag)
{
  /* Handle the width attribute for various tags */
  
  int quoted = 0;
  long width;
  
  debug (debug_parse_attribs, "handle_width: Entering pointing at %ld:%s...",
	 *index, get_bit(document,index) ? foo : "NULL");
  (*index) += 5; /*Skip WIDTH */
  debug (debug_parse_attribs, "handle_width: Skipped WIDTH, now pointing at %ld:%s...",
	 *index, get_bit(document,index) ? foo : "NULL");
  if (skip_equals(document, index) == -1 )
    return;
  if ((document[*index]=='"'))
    {
      debug (debug_parse_attribs, "handle_width: Found \".");
      (*index)++;
      quoted = 1;
    }
  if (isdigit(document[*index]))
    width = atoi (&(document[*index])); /* Standard library function, need to check if thi
					 * is robust enough for us */
  
  /* Skip until we find a space or end of tag if we're not
   * quoted or until we find a close quote if we're quoted
   */
  while (((document[*index] != ' ' && document[*index] != '>') && !quoted) ||
	 ((document[*index] != '"') && quoted))
    (*index)++;
  if (document[*index] == '"')
    (*index)++;
  if (tag == PRE || tag == HRULE || tag == IMAGE || tag == APPLET)
    {
      debug ( debug_parse_attribs, "handle_width: Setting width to %ld.", width);
      node->physical_attributes.width = width;
    }
  debug (debug_parse_attribs, "handle_width: Leaving pointing at %ld:%s...",
	 *index, get_bit(document,index) ? foo : "NULL");
}



void handle_height (const char *document, long *index, neve_in_chains *node, int tag)
{
  /* Essentially the same as handle_width, see above for commentary */
  int quoted = 0;
  long height;
  
  debug (debug_parse_attribs, "handle_height: Entering pointing at %ld:%s...",
	 *index, get_bit(document,index) ? foo : "NULL");
  (*index) += 6; /*Skip HEIGHT */
  debug (debug_parse_attribs, "handle_width: Skipped HEIGHT, now pointing at %ld:%s...",
	 *index, get_bit(document,index) ? foo : "NULL");
  if (skip_equals(document, index) == -1 )
    return;
  if (document[*index]=='"')
    {
      debug ( debug_parse_attribs, "handle_height: Found a \".");
      quoted = 1;
      (*index)++;
    }
  if (isdigit(document[*index]))
    height = atoi (&(document[*index]));
  while (((document[*index] != ' ' && document[*index] != '>') && !quoted) ||
	 (document[*index] != '"' && quoted))
    (*index)++;
  if (document[*index] == '"')
    (*index)++;
  if (tag == IMAGE || tag == APPLET)
    {
      node->physical_attributes.height = height;
      debug (debug_parse_attribs, "handle_height: Setting height to %ld", height);
    }
  debug (debug_parse_attribs, "handle_width: Leaving pointing at %ld:%s...",
	 *index, get_bit(document,index) ? foo : "NULL");
}



void handle_bgcolor (BrowserWindow *window, const char *document, long *index, int tag)
{
  /* parse the color="#xxxxxx" structure, should also parse the common
   * color names, but erm doesn't yet.  
   */
  colour_type value = {-1,-1,-1};
  (*index)+=7; /* Skip BGCOLOR */
  if (skip_equals (document, index) == -1)
    return;
  if (document[*index] == '"')
    (*index)++;
  if (document[*index] == '#') 
    {
      (*index)++; /* Skip the # */
      if ((value.red = convert_hex (document, index)) == -1)
	return;
      debug ( debug_parse_attribs, "handle_colour: Red value = %d", value.red);  
      if ((value.green = convert_hex (document, index)) == -1)
	return;
      debug ( debug_parse_attribs, "handle_colour: Green value = %d", value.green);
      if ((value.blue = convert_hex (document, index)) == -1)
	return;
      debug ( debug_parse_attribs, "handle_colour: Blue value = %d", value.blue);
      if (document[*index] =='"')
	(*index)++;
      if (tag == BODY) /* Doesn't make any sense for any other tag */
	{
	  debug ( debug_parse_attribs, "handle_bgcolor: calling lisa_set_colours with %d.%d.%d, NULL, NULL", value.red, value.green, value.blue);
	  lisa_set_colours (GTK_LISA(window->lisa), &value,  NULL, NULL, NULL ); 
 	} 
    }
  else
    {
      ; /* Should check the named colors here */
    }
}



void handle_text (const char *document, long *index, int tag, physical_attribute_types *pattribs)
{
  /* parse the color="#xxxxxx" structure, should also parse the common
   * color names, but erm doesn't yet.  
   */
  colour_type value = {-1,-1,-1};
  (*index)+=4; /* Skip TEXT */
  if (skip_equals (document, index) == -1)
    return;
  if (document[*index] == '"')
    (*index)++;
  if (document[*index] == '#') 
    {
      (*index)++; /* Skip the # */
      if ((value.red = convert_hex (document, index)) == -1)
	return;
      debug ( debug_parse_attribs, "handle_colour: Red value = %d", value.red);  
      if ((value.green = convert_hex (document, index)) == -1)
	return;
      debug ( debug_parse_attribs, "handle_colour: Green value = %d", value.green);
      if ((value.blue = convert_hex (document, index)) == -1)
	return;
      debug ( debug_parse_attribs, "handle_colour: Blue value = %d", value.blue);
      if (document[*index] =='"')
	(*index)++;
      if (tag == BODY) /* Doesn't make any sense for any other tag */
	{
	  debug ( debug_parse_attribs, "handle_text: Setting font colour to %d.%d.%d", value.red, value.green, value.blue);
	  pattribs->draw_colour = value;
 	} 
    }
  else
    {
      ; /* Should check the named colors here */
    }
}


void handle_link (BrowserWindow *window, const char *document, long *index, int tag)
{
  /* parse the color="#xxxxxx" structure, should also parse the common
   * color names, but erm doesn't yet.  
   */
  colour_type value = {-1,-1,-1};
  (*index)+=4; /* Skip LINK */
  if (skip_equals (document, index) == -1)
    return;
  if (document[*index] == '"')
    (*index)++;
  if (document[*index] == '#') 
    {
      (*index)++; /* Skip the # */
      if ((value.red = convert_hex (document, index)) == -1)
	return;
      debug ( debug_parse_attribs, "handle_colour: Red value = %d", value.red);  
      if ((value.green = convert_hex (document, index)) == -1)
	return;
      debug ( debug_parse_attribs, "handle_colour: Green value = %d", value.green);
      if ((value.blue = convert_hex (document, index)) == -1)
	return;
      debug ( debug_parse_attribs, "handle_colour: Blue value = %d", value.blue);
      if (document[*index] =='"')
	(*index)++;
      if (tag == BODY) /* Doesn't make any sense for any other tag */
	{
	  debug ( debug_parse_attribs, "handle_text: Setting link colour to %d.%d.%d", value.red, value.green, value.blue);
	  lisa_set_colours (GTK_LISA(window->lisa), NULL, &value, NULL, NULL);
 	} 
    }
  else
    {
      ; /* Should check the named colors here */
    }
}


void handle_alink (BrowserWindow *window, const char *document, long *index, int tag)
{
  /* parse the color="#xxxxxx" structure, should also parse the common
   * color names, but erm doesn't yet.  
   */
  colour_type value = {-1,-1,-1};
  (*index)+=5; /* Skip ALINK */
  if (skip_equals (document, index) == -1)
    return;
  if (document[*index] == '"')
    (*index)++;
  if (document[*index] == '#') 
    {
      (*index)++; /* Skip the # */
      if ((value.red = convert_hex (document, index)) == -1)
	return;
      debug ( debug_parse_attribs, "handle_colour: Red value = %d", value.red);  
      if ((value.green = convert_hex (document, index)) == -1)
	return;
      debug ( debug_parse_attribs, "handle_colour: Green value = %d", value.green);
      if ((value.blue = convert_hex (document, index)) == -1)
	return;
      debug ( debug_parse_attribs, "handle_colour: Blue value = %d", value.blue);
      if (document[*index] =='"')
	(*index)++;
      if (tag == BODY) /* Doesn't make any sense for any other tag */
	{
	  debug ( debug_parse_attribs, "handle_text: Setting alink colour to %d.%d.%d", value.red, value.green, value.blue);
	  lisa_set_colours (GTK_LISA(window->lisa), NULL, NULL, NULL, &value);
 	} 
    }
  else
    {
      ; /* Should check the named colors here */
    }
}


void handle_vlink (BrowserWindow *window, const char *document, long *index, int tag)
{
  /* parse the color="#xxxxxx" structure, should also parse the common
   * color names, but erm doesn't yet.  
   */
  colour_type value = {-1,-1,-1};
  (*index)+=5; /* Skip VLINK */
  if (skip_equals (document, index) == -1)
    return;
  if (document[*index] == '"')
    (*index)++;
  if (document[*index] == '#') 
    {
      (*index)++; /* Skip the # */
      if ((value.red = convert_hex (document, index)) == -1)
	return;
      debug ( debug_parse_attribs, "handle_colour: Red value = %d", value.red);  
      if ((value.green = convert_hex (document, index)) == -1)
	return;
      debug ( debug_parse_attribs, "handle_colour: Green value = %d", value.green);
      if ((value.blue = convert_hex (document, index)) == -1)
	return;
      debug ( debug_parse_attribs, "handle_colour: Blue value = %d", value.blue);
      if (document[*index] =='"')
	(*index)++;
      if (tag == BODY) /* Doesn't make any sense for any other tag */
	{
	  debug ( debug_parse_attribs, "handle_text: Setting vlink colour to %d.%d.%d", value.red, value.green, value.blue);
	  lisa_set_colours (GTK_LISA(window->lisa), NULL, NULL, &value, NULL);
 	} 
    }
  else
    {
      ; /* Should check the named colors here */
    }
}



void parse_attribs (BrowserWindow *window, const char *document,long *index, neve_in_chains *neve_node, 
		    int tag, int *current_basefont, physical_attribute_types *current_pattribs, 
		    logical_attribute_types *current_lattribs, char *current_href)
{
  /* This is the main function which loops along the tags parametes, mathching
   * them if possible and calling the relevant functions.
   * Should be pointing at the space after the tag name or the closing > on entry 
   */
  
  boolean font_change_implemented = FALSE; /* Flag to record if we've changed any FONT attributes
				    * in this invocation so we know whether to create a new
				    * font_stack node or just add the information to the
				    * top node */
  if (document[*index] == '\0') /* Check for end of doc */
    return;
  
  if (document[*index] == '>')
    {
      debug (debug_parse_attribs, "parse_attribs: Found end of tag.");
      (*index)++;
      return;
    }
  
  debug (debug_parse_attribs, "parse_attribs: Entering main loop pointing at %ld:%s...",
	 *index, get_bit(document, index) ? foo : "NULL");
  
  while (document[*index] != '\0' && document[*index] != '>') 
    {
      /* Skip whitespace before the attribute name */
      while (document[*index] != '\0' && document[*index] == ' ')
	(*index)++;
      /* Try to match the attribute name */
      if (strncasecmp(&document[*index], "alt", 3) == 0)
	{
	  debug (debug_parse_attribs, "parse_attribs: calling handle_alt ...");
	  handle_alt(document, index, neve_node, tag);
	}
      else if (strncasecmp(&document[*index], "href", 4) == 0)
	{
	  debug (debug_parse_attribs, "parse_attribs: calling handle_href ...");
	  handle_href(document, index, tag, current_href, window);
	  debug (debug_parse_attribs, "chref=%s", current_href);
	}
      else if (strncasecmp(&document[*index], "src", 3) == 0)
	{
	  debug (debug_parse_attribs, "parse_attribs: calling handle_src ...");
	  handle_src(document, index, neve_node, tag);
	}
      else if (strncasecmp(&document[*index], "align", 5) == 0)
	{
	  debug (debug_parse_attribs, "parse_attribs: calling handle_align ...");
	  handle_align(document, index, neve_node, tag, current_pattribs);
	}
      else if (strncasecmp(&document[*index], "size", 4) == 0)
	{
	  debug (debug_parse_attribs, "parse_attribs: calling handle_size ...");
	  handle_size(document, index, neve_node, tag, current_basefont, 
		      &font_change_implemented, current_pattribs);
	}
      else if (strncasecmp(&document[*index], "width", 5) == 0)
	{
	  debug (debug_parse_attribs, "parse_attribs: calling handle_width ...");
	  handle_width(document, index, neve_node, tag);
	}
      else if (strncasecmp(&document[*index], "height", 6) == 0)
	{
	  debug (debug_parse_attribs, "parse_attribs: calling handle_height ...");
	  handle_height(document, index, neve_node, tag);
	}    
      else if (strncasecmp(&document[*index], "type", 4) == 0)
	{
	  debug (debug_parse_attribs, "parse_attribs: calling handle_type ...");
	  handle_type (document, index, current_lattribs, tag);
	}
      else if (strncasecmp(&document[*index], "color", 5) == 0)
	{
	  debug (debug_parse_attribs, "parse_attribs: calling handle_colour ...");
	  handle_colour(document, index, tag, &font_change_implemented, current_pattribs);
	}
      else if (strncasecmp(&document[*index], "bgcolor", 7) == 0)
	{
	  debug (debug_parse_attribs, "parse_attribs: calling handle_bgcolor ...");
	  handle_bgcolor(window, document, index, tag);
	}
      else if (strncasecmp(&document[*index], "text", 4) == 0)
	{
	  debug (debug_parse_attribs, "parse_attribs: calling handle_text ...");
	  handle_text(document, index, tag, current_pattribs);
	}
      else if (strncasecmp(&document[*index], "link", 4) == 0)
	{
	  debug (debug_parse_attribs, "parse_attribs: calling handle_link ...");
	  handle_link (window, document, index, tag);
	}
      else if (strncasecmp(&document[*index], "vlink", 5) == 0)
	{
	  debug (debug_parse_attribs, "parse_attribs: calling handle_vlink ...");
	  handle_vlink (window, document, index, tag);
	}
      else if (strncasecmp(&document[*index], "alink", 6) == 0)
	{
	  debug (debug_parse_attribs, "parse_attribs: calline handle_alink ...");
	  handle_alink (window, document, index, tag);
	}
    else
      {
	debug (debug_parse_attribs, 
	       "parse-attribs: Don't recognize this attrib, try and skip from %ld:%s...",
	       *index, get_bit (document, index) ? foo : "NULL");
	handle_unrecognized(document, index);
	debug (debug_parse_attribs, 
	       "parse_attribs: Finished the unrecognized one, now pointing at %ld:%s...",
	       *index, get_bit (document, index) ? foo : "NULL");
      }
  }
  
  if (document[*index] != '\0')
    (*index)++; /* Skip past the last > in the tag */

  debug (debug_parse_attribs,"parse_attribs: Leaving pointing at %ld:%s...",
	 *index, get_bit (document, index) ? foo : "NULL");
  
  return;
}  




  

