/* YAWN, Yet Another Web Navigator */
/* Copyright (C) 1998, James Troup <jjtroup@comp.brad.ac.uk>
 *                     Lee Willis <lwillis@comp.brad.ac.uk> */

/* This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "utils.h"
#include "scanner.h"
#include "tagdefs.h"
#include "latin.h"

/* scanner.c contains the functions which scan the document and either return
 * text or a tag id number as defined in tagdefs.h
 */

#define MT(TAG) (my_tagcmp (text+*i,TAG,strlen(TAG)) == 0)

int my_tagcmp (char *text1,const char *text2, int length) 
{ 
  /* Do a case-insensitive compare of two strings and check if
   * there is whitespace or a > after the desired length so that
   * we don't match LI when we've really got LIST
   */
  int result;
  
  result = strncasecmp (text1, text2, length);
  if (result == 0)
    if ((is_whitespace(text1[length])) || (text1[length] == '>'))
      return 0;
    else
      return -1;
  else
    return -1;
}


void fold(char *buffer)
{
  /* Folds whitespace in a text string as defined in the HTML
   * standard. E.g sequences of more than one character of whitespace
   * are treated as a single space character, and all new line and tab
   * characters are also replaced with a single space character
   */

  int from = 0;
  int to = 0;

  debug (debug_scanner, "fold: Called pointing at $$%s$$.", buffer);
  while (buffer[from] != '\0')
    {
      if (is_whitespace(buffer[from]))
	  buffer[from] = ' ';
      buffer[to] = buffer[from];
      if (!(to>0 && buffer[to-1] == ' ' && buffer[from] ==' '))
	to++;
      from++;
    }
  if (buffer[from] == '\0')
    buffer[to] = '\0';
  debug (debug_scanner, "fold: Leaving pointing at $$%s$$.", buffer);
}



int scan (char *document, long *index, char *text, int preformatted, int prev_tag)
{
  /* Scans the text to find either text or a tag */

  int buffer_p = 0; /* Pointer to show where to place a text character
		     * if we find it. Offset into the buffer text */

  debug (debug_scanner, "scan: Entered pointing at %c", document[*index]);
  
  if (!preformatted && prev_tag < CLOSETAG && prev_tag != TEXT) /* Fold newlines after an open tag */
    if (*index>0 && document[*index] != '\0') 
      while (document[*index]=='\n')
	{
	  debug (debug_scanner, "scan: skipping \\n after open tag thingy ...");
	  (*index)++; 
	}
  
  while (document[*index] != '\0')
    {
      if (document[*index] != '<')
	{
	  if (buffer_p < BUF_SIZE - 1) 
	    {
	      /* Add a character to the text buffer */
	      text[buffer_p] = document[*index];
	      buffer_p++;
	      (*index)++;
	    }
	  else /* Buffer full, ensure it's NULL terminated and return... */
	    {
	      debug (debug_scanner, "scan: Buffer full, returning text.");
	      text[buffer_p]='\0';
	      parse_latin_entities(text); /* Parse &xxx; entities */
	      debug (debug_scanner, "scan: returning $$%s$$.", text);
	      return TEXT;
	    }
	}
      else
	{ /* We've found what might be a tag, terminate the text
	   * buffer if we have one and return */
	  if (buffer_p)
	    { 
	      debug (debug_scanner, "scan: Returning text in buffer ...");
	      debug (debug_scanner, "scan: buffer_p is %d", buffer_p);
	      text[buffer_p] = '\0'; /* NULL terminate the text buffer */
	      parse_latin_entities(text); /* Parse the Latin entities */
	      debug (debug_scanner, "scan: returning $$%s$$.", text);
	      return TEXT; 
	    } 
	  else 
	    { 
	      debug (debug_scanner, "scan: Possibly found a tag ...");
	      (*index)++;
	      return parse_tag (document, index, text); 
	    } 
	} 
    }
  if (buffer_p)
    { 
      debug (debug_scanner, "scan: Returning text in buffer ...");
      debug (debug_scanner, "scan: buffer_p is %d", buffer_p);
      text[buffer_p] = '\0'; /* NULL terminate the text buffer */
      parse_latin_entities(text); /* Parse the Latin entities */
      debug (debug_scanner, "scan: returning $$%s$$.", text);
      return TEXT; 
    }
  else
    return END_OF_DOC;
}






int parse_tag(char *text, long *i, char *text_buffer)
{

  /* Should be pointing at the first non-ws char after the < */
  int tag = UNRECOGNIZED;

  debug (debug_scanner, "parse_tag: Pointing at %c", text[*i]);
  
  if ((text[*i] >= 'a' && text[*i] <= 'z') || 
      (text[*i] >= 'A' && text[*i] <= 'Z') ||
      (text[*i] == '/') ||
      (text[*i] == '!'))
    tag=match_tag (text,i);
  else
    {  /* text[*i] not an alphabetic character or !/ 
	  therefore this isn't a proper tag (SGML def'n)*/
      
      /* What should we do here??  Kind of mucks everything up for a
       * while We could I s'pose plonk it in the buffer, and return
       * TEXT; I have done, still not convinced it's right etc.
       */
      
      text_buffer[0] = '<';
      text_buffer[1] = '\0';
      debug (debug_scanner, "scan: returning $$%s$$.", text);
      return TEXT;
    }
  
  /* parse attribs */
  
  return tag;

}



void parse_comment(char *text, long *i)
{
  int count = 1;
  
  debug (debug_scanner, "parse_comment entered pointing at %c%c%c%c", text[*i], text[*i+1], text[*i+2], text[*i+3]);
  while (text[*i]  && count) 
    {
      if (text[*i] == '<') 
 	count++; 
      else if (text[*i] == '>') 
 	count--; 
      (*i)++;
    }
  debug (debug_scanner, "parse_comment leaving pointing at %c%c%c%c", text[*i], text[*i+1], text[*i+2], text[*i+3]);
}



int match_tag (char *text, long *i)
{

  int tag = 0; /* Tag number to be returned */

  debug (debug_scanner, "match_tag: pointing at %c", text[*i]);

  if ((text[*i]) == '/') {
    tag = CLOSETAG;
    (*i)++;
  }
  
  if (text[*i] == '!') { /* Nasty if we have </! foobar> it'll call a comment */
    
    /* Code to handle comments goes here!
     * As soon as I work out what it should be!
     * Actually this is probably OK, but it does
     * really need checking over */

    (*i)++; /* Skip the ! */
    tag = COMMENT;
    parse_comment (text, i);
  }
  
  else {
    
    /* Note: Currently tags which can be subsets of other tags ie. A is a subset of ADDRESS, LI is 
     * a subset of LISTING If you see what I mean must be after their longer counterparts to avoid
     * matching the wrong tag, I'll fix this by checking for a space somewhere or other ...
     * Frankly this sucks ... */
    /* Fixed I think see #define of MT for an explanation */

    if MT("address")
      tag |= ADDRESS; 
    else if MT("a")
      tag |= ANCHOR;
    else if MT("basefont")
      tag |= BASEFONT;
    else if MT("base")
      tag |= BASE;
    else if MT("big")
      tag |= BIG;
    else if MT("blockquote")
      tag |= BLOCKQUOTE;
    else if MT("br")
      tag |= BR;
    else if MT("body")
      tag |= BODY;
    else if MT("b")
      tag |= BOLD;
    else if MT("center")
      tag |= CENTER;
    else if MT("cite")
      tag |= CITE;
    else if MT("code")
      tag |= CODE;
    else if MT("dir")
      tag |= DIR;
    else if MT("dd")
      tag |= DD;
    else if MT("dl")
      tag |= DL;
    else if MT("dt")
      tag |= DT;
    else if MT("em")
      tag |= ITALIC;   /* Just call it italic since we don't have CSS */
    else if MT("font")
      tag |= FONT;
    else if MT("h1")
      tag |= H1;
    else if MT("h2")
      tag |= H2;
    else if MT("h3")
      tag |= H3;
    else if MT("h4")
      tag |= H4;
    else if MT("h5")
      tag |= H5;
    else if MT("h6")
      tag |= H6;
    else if MT("hr")
      tag |= HRULE;
    else if MT("head")
      tag |= HEAD;
    else if MT("html")
      tag |= HTML;
    else if MT("img")
      tag |= IMAGE;
    else if MT("i")
      tag |= ITALIC;
    else if MT("kbd")
      tag |= KBD;
    else if MT("listing")
      tag |= LISTING;
    else if MT("li")
      tag |= LIST;
    else if MT("menu")
      tag |= MENU;
    else if MT("ol")
      tag |= OLIST;
    else if MT("plaintext")
      tag |= PLAINTEXT;
    else if MT("pre")
      tag |= PRE;
    else if MT("p")
      tag |= PARAGRAPH;
    else if MT("script")
      tag |= SCRIPT;
    else if MT("small")
      tag |= SMALL;
    else if MT("strike")
      tag |= STRIKE;
    else if MT("strong")
      tag |= BOLD; /* See <EM> */
    else if MT("style")
      tag |= STYLE;
    else if MT("sub")
      tag |= SUBSCRIPT;
    else if MT("sup")
      tag |= SUPSCRIPT;
    else if MT("table")
      tag |= TABLE;
    else if MT("tt")
      tag |= TELETYPE;
    else if MT("td")
      tag |= TD;
    else if MT("title")
      tag |= TITLE;
    else if MT("tr")
      tag |= TR;
    else if MT("ul")
      tag |= ULIST;
    else if MT("underline")
      tag |= UNDERLINE;
    else if MT("xmp")
      tag |= XMP;
    
    /* Should leave pointing at first non-blank char after the tag name so we can either
     * parse attributes if applicable or just skip to end of tag Ron */

    while ((!is_whitespace(text[*i])) && text[*i] != '>')
      (*i)++;
    
  }
  debug (debug_scanner, "match_tag: Returning %d", tag);
  return tag;
}
